#!/usr/bin/env python

"""
This module is a part of:
WAAFLE, a [W]orkflow to [A]nnotate [A]ssemblies and [F]ind [L]GT [E]vents

Copyright (c) 2019 Harvard T.H. Chan School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function # Python 2.7+ required
import os
import sys
import csv
import argparse
import re
from collections import Counter

import numpy as np

from waafle import utils as wu

# ---------------------------------------------------------------
# description
# ---------------------------------------------------------------

description = wu.describe( """
{SCRIPT}: Generate gene-gene junction stats for contig QC

This script maps reads to contigs (or analyzes an existing mapping)
to find support for gene-gene junctions. A junction is supported if
there are mate-pairs that span the junction. As this may be infeasible
for larger junctions (e.g. those exceeding a typical insert size of 300 nt),
a junction can also be supported by having a reasonable average coverage
relative to its flanking genes.
""" )

# ---------------------------------------------------------------
# output formats
# ---------------------------------------------------------------

c_formats = {}

c_formats["site_hits"] = """
contig
mean
stdev
depths
"""

c_formats["gene_hits"] = """
contig
gene1
gene2
hits
"""

c_formats["junctions"] = """
contig
gene1
gene2
len_gene1
len_gene2
gap
junction_hits
coverage_gene1
coverage_gene2
coverage_junction
ratio
"""

for name, items in c_formats.items( ):
    c_formats[name] = [k for k in items.split( "\n" ) if k != ""]

# ---------------------------------------------------------------
# cli
# ---------------------------------------------------------------

def get_args( ):

    parser = argparse.ArgumentParser(
        description=description, 
        formatter_class=argparse.RawTextHelpFormatter,
        )

    g = parser.add_argument_group( "required inputs" )
    g.add_argument( 
        "contigs",
        help="contigs file (fasta format)",
        )
    g.add_argument( 
        "gff",
        help="GFF file for provided contigs",
        )

    g = parser.add_argument_group( "provide paired reads or a .sam file" )
    g.add_argument( 
        "--reads1",
        metavar="<path>",
        help="sequencing reads (mate-1)",
        )
    g.add_argument( 
        "--reads2",
        metavar="<path>",
        help="sequencing reads (mate-2)",
        )
    g.add_argument( 
        "--sam",
        metavar="<path>",
        help="sam file (from existing alignment)",
        )

    g = parser.add_argument_group( "output options" )
    g.add_argument( 
        "--tmpdir",
        default=".",
        metavar="<path>",
        help="where to place temp outputs\n[default: ./]",
        )
    g.add_argument( 
        "--outdir",
        default=".",
        metavar="<path>",
        help="where to place main outputs\n[default: ./]",
        )
    g.add_argument( 
        "--basename",
        metavar="<str>",
        help="basename for output files\n[default: <derived from input>]",
        )
    g.add_argument( 
        "--write-detailed-output",
        action="store_true",
        help="write out coverage values for all sites and all junctions\n[default: off]",
        )

    g = parser.add_argument_group( "filtering parameters" )
    g.add_argument(
        "--min-overlap-sites",
        type=int,
        default=25,
        metavar="<int>",
        help="minimum nucleotide overlap for counting a read-gene hit\n[default: 25]",
        )

    g = parser.add_argument_group( "bowtie2 options" )
    g.add_argument( 
        "--bowtie2-build",
        default="bowtie2-build",
        metavar="<path>",
        help="path to bowtie2-build\n[default: $PATH]",
        )
    g.add_argument( 
        "--bowtie2",
        default="bowtie2",
        metavar="<path>",
        help="path to bowtie2\n[default: $PATH]",
        )
    g.add_argument( 
        "--threads",
        type=int,
        default=1,
        metavar="<int>",
        help="number of threads for bowtie2 steps\n[default: 1]",
        )
    g.add_argument( 
        "--resume",
        action="store_true",
        help="if set, use existing .index and/or .sam if found\n[default: off]",
        )

    args = parser.parse_args( )
    return args

# ---------------------------------------------------------------
# utils for running bowtie2
# ---------------------------------------------------------------

def bowtie2_build( p_bowtie2_build=None, p_contigs=None, p_index=None, 
                   args=None, ):
    alias = {
        "PROG":    p_bowtie2_build,
        "CONTIGS": p_contigs, 
        "INDEX":   p_index, 
        }
    if args.resume and os.path.exists( p_index + ".1.bt2" ):
        wu.say( "RESUMING: The index <{INDEX}> already exists.".format( **alias ) )
    else:
        wu.say( "Indexing <{CONTIGS}> to <{INDEX}>.".format( **alias ) )
        command = [
            "{PROG}",
            "{CONTIGS}",
            "{INDEX}",
            ]
        command = " ".join( command )
        command = command.format( **alias )
        os.system( command )
        wu.say( "Build complete." )
    return None

def bowtie2_align( p_bowtie2=None, p_reads1=None, p_reads2=None, 
                   p_index=None, p_sam=None, args=None, ):
    alias = {
        "PROG":    p_bowtie2,
        "READS1":  p_reads1,
        "READS2":  p_reads2,
        "INDEX":   p_index,
        "SAM":     p_sam,
        "THREADS": args.threads,
        }
    if args.resume and os.path.exists( p_sam ):
        wu.say( "RESUMING: A sam mapping <{SAM}> already exists.".format( **alias ) )
    else:
        wu.say( "Performing bowtie2 alignment." )
        command = [
            "{PROG}",
            "-x {INDEX}",
            "-1 {READS1}",
            "-2 {READS2}",
            "-S {SAM}",
            "--threads {THREADS}",
            "--no-mixed",
            "--no-discordant",
            ]
        command = " ".join( command )
        command = command.format( **alias )
        os.system( command )
        wu.say( "Alignment complete." )
    return None

# ---------------------------------------------------------------
# utils for parsing SAM/GFF comparison
# ---------------------------------------------------------------

def concordant_hits( p_sam=None, ):
    counter = 0
    mate1 = None
    mate2 = None
    for hit in wu.iter_sam_hits( p_sam ):
        # progress
        counter += 1
        if counter % int( 1e5 ) == 0:
            wu.say( "  SAM alignments processed: {:.1f}M".format( counter / 1e6 ) )
        # weave
        mate1 = mate2
        mate2 = hit
        # edge case
        if mate1 is None:
            continue
        # not a mate pair
        elif mate1.qseqid != mate2.qseqid:
            continue
        # not concordant
        elif mate1.sseqid != mate2.sseqid:
            continue
        # good pair
        else:
            yield [mate1, mate2]

def find_hit_loci( mate1=None, mate2=None, loci=None, args=None ):
    hits = set( )
    for L in loci:
        a1, a2 = L.start, L.end
        for read in [mate1, mate2]:
            b1, b2 = read.sstart, read.send
            overlap = wu.calc_overlap( a1, a2, b1, b2, normalize=False )
            if overlap >= args.min_overlap_sites:
                hits.add( L.code )
    return hits

# ---------------------------------------------------------------
# utils for evaluating a contig
# ---------------------------------------------------------------

def evaluate_contig( loci=None, coverage=None, gene_hits=None, args=None, ):
    rowdicts = []
    loci = sorted( loci, key=lambda x: x.start )
    for i in range( len( loci ) - 1 ):
        rowdict = {}
        L1 = loci[i]
        L2 = loci[i+1]
        rowdict["gene1"]             = my_code1 = L1.code
        rowdict["gene2"]             = my_code2 = L2.code
        rowdict["len_gene1"]         = len( L1 )
        rowdict["len_gene2"]         = len( L2 )
        rowdict["gap"]               = my_gap = L2.start - L1.end - 1
        # check hits
        rowdict["junction_hits"]     = my_hits = gene_hits.get( (my_code1, my_code2), 0 )
        # check coverage (note: base-0 start and pythonic end)
        rowdict["coverage_gene1"]    = my_cov1 = np.mean( coverage[L1.start-1:L1.end] )
        rowdict["coverage_gene2"]    = my_cov2 = np.mean( coverage[L2.start-1:L2.end] )
        # define junction coverage as 0 if there's no gap
        my_covj = 0.0 if my_gap <= 0 else np.mean( coverage[L1.end-1:L2.start] )
        rowdict["coverage_junction"] = my_covj
        # note: pseudocount to avoid /0
        my_mean = np.mean( [my_cov1, my_cov2] )
        rowdict["ratio"]             = my_ratio = my_covj / ( my_mean + 1e-6 )
        rowdicts.append( rowdict )
    return rowdicts

# ---------------------------------------------------------------
# utils for output formatting
# ---------------------------------------------------------------

def write_detailed_output( basename=None, outdir=None,
                           contig_coverage=None, contig_hits=None, ):

    # first file can get pretty big, hence gzip
    p_site_hits = wu.name2path( basename, outdir, ".site_hits.tsv.gz" )
    p_gene_hits = wu.name2path( basename, outdir, ".gene_hits.tsv" )

    # write: site_hits
    wu.say( "Writing site hits." )
    with wu.try_open( p_site_hits, "w" ) as fh:
        wu.write_rowdict( 
            format=c_formats["site_hits"], 
            file=fh, )
        for c in sorted( contig_coverage ):
            depths = contig_coverage[c]
            rowdict = {
                "contig" : c,
                "mean"   : np.mean( depths ),
                "stdev"  : np.std( depths ),
                "depths" : " ".join( ["{:.0f}".format( k ) for k in depths] ),
                }
            wu.write_rowdict( 
                rowdict=rowdict, 
                format=c_formats["site_hits"], 
                file=fh, )

    # write: gene_hits
    wu.say( "Writing gene-pair hits." )
    with wu.try_open( p_gene_hits, "w" ) as fh:
        wu.write_rowdict( 
            format=c_formats["gene_hits"], 
            file=fh, )
        for c in sorted( contig_hits ):
            for code1, code2 in sorted( contig_hits[c] ):
                if code2 > code1:
                    continue
                value = contig_hits[c][(code1, code2)]
                rowdict = {
                    "contig" : c,
                    "gene1"  : code1,
                    "gene2"  : code2,
                    "hits"   : value,
                    }
                wu.write_rowdict( 
                    rowdict=rowdict, 
                    format=c_formats["gene_hits"], 
                    file=fh, )

# ---------------------------------------------------------------
# main
# ---------------------------------------------------------------

def main( ):

    # begin
    args = get_args( )
    p_contigs = args.contigs
    p_gff = args.gff

    # define files
    p_outdir = args.outdir
    p_tmpdir = args.tmpdir
    basename = args.basename
    if basename is None:
        basename = wu.path2name( p_contigs )
    p_index     = wu.name2path( basename, p_tmpdir, ".index" )
    p_sam       = wu.name2path( basename, p_tmpdir, ".sam" )
    p_junctions = wu.name2path( basename, p_outdir, ".junctions.tsv" )

    # alignment workflow
    if args.sam is not None:
        p_sam = args.sam
        wu.say( "Using specified SAM file:", p_sam )
    elif args.reads1 is not None and args.reads2 is not None:
        # build process
        bowtie2_build( 
            p_bowtie2_build=args.bowtie2_build,
            p_contigs=args.contigs,
            p_index=p_index,
            args=args,
            )
        # alignment process
        bowtie2_align(
            p_bowtie2=args.bowtie2,
            p_reads1=args.reads1, 
            p_reads2=args.reads2,
            p_index=p_index,
            p_sam=p_sam,
            args=args,
            )
    else:
        wu.die( "Must provide READS or SAM file." )

    # load contig data
    wu.say( "Loading contig lengths." )
    contig_lengths = wu.read_contig_lengths( p_contigs )
    contig_coverage = {}
    for name, length in contig_lengths.items( ):
        contig_coverage[name] = np.zeros( length )
    wu.say( "Loading contig gene coordinates." )
    contig_loci = {}
    for name, loci in wu.iter_contig_loci( p_gff ):
        contig_loci[name] = loci
    contig_hits = {}

    # post-processing workflow
    wu.say( "Processing SAM file." )
    for mate1, mate2 in concordant_hits( p_sam ):
        contig = mate1.sseqid
        inner = contig_hits.setdefault( contig, Counter( ) )
        # update pers-site coverage (note: base-0 start and pythonic end)
        coords = [mate1.sstart, mate1.send, mate2.sstart, mate2.send]
        L = min( coords ) - 1
        R = max( coords ) - 1
        contig_coverage[contig][L:R+1] += 1
        # find hit loci
        hits = find_hit_loci( 
            mate1=mate1, 
            mate2=mate2, 
            loci=contig_loci.get( contig, [] ),
            args=args,
            )
        # attach self counts
        for code in hits:
            inner[(code, code)] += 1
        # attach pair counts (note: symmetric storage for safer lookup)
        for code1 in hits:
            for code2 in hits:
                if code1 != code2:
                    inner[(code1, code2)] += 1

    # detailed output?
    if args.write_detailed_output:
        write_detailed_output(
            basename=basename,
            outdir=p_outdir,
            contig_coverage=contig_coverage,
            contig_hits=contig_hits,
            )

    # write junction report
    wu.say( "Writing junction report." )
    with wu.try_open( p_junctions, "w" ) as fh:
        wu.write_rowdict( 
            format=c_formats["junctions"], 
            file=fh, )
        for c in sorted( contig_lengths ):
            rowdicts = evaluate_contig( 
                loci=contig_loci.get( c, [] ),
                coverage=contig_coverage[c], 
                gene_hits=contig_hits.get( c, {} ),
                args=args,
                )
            for rowdict in rowdicts:
                rowdict["contig"] = c
                wu.write_rowdict( 
                    rowdict=rowdict, 
                    format=c_formats["junctions"], 
                    file=fh, )

    # end
    wu.say( "Finished successfully." )

if __name__ == "__main__":
    main( )



#!/usr/bin/env python

"""
This module is a part of:
WAAFLE, a [W]orkflow to [A]nnotate [A]ssemblies and [F]ind [L]GT [E]vents

Copyright (c) 2019 Harvard T.H. Chan School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function # Python 2.7+ required

import os
import sys
import re
import argparse
from collections import Counter

import numpy as np

from waafle import utils as wu
from waafle.waafle_genecaller import attach_shared_args

# ---------------------------------------------------------------
# description
# ---------------------------------------------------------------

description = wu.describe( """
{SCRIPT}: Step 2 in the WAAFLE pipeline

Merges blast hits into genes on contigs-of-interest. Uses corresponding
taxonomy file, and the WAAFLE algorithm, to identify contigs that are
best explained by a single clade vs. a pair of clades. The latter events
correpond to putative LGTs.
""" )

# ---------------------------------------------------------------
# constants
# ---------------------------------------------------------------

c_eps                = 1e-6
c_precision          = 3
c_annotation_prefix  = "ANNOTATIONS:"
c_missing_annotation = "None"
c_empty_field        = "--"
c_delim0             = "\t"
c_delim1             = "; "
c_delim2             = "|"
c_delim3             = ":"
# note: "A" and "B" synchars are hardcoded
c_synchar_ambiguous  = "*"
c_synchar_ignored    = "~"
c_synchar_error      = "!"

# ---------------------------------------------------------------
# output formats
# ---------------------------------------------------------------

c_formats = {}

c_formats["lgt"] = """
contig_name
call
contig_length
min_max_score
avg_max_score
synteny
direction
clade_A
clade_B
lca
melded_A
melded_B
taxonomy_A
taxonomy_B
loci
"""

c_formats["no_lgt"] = """
contig_name
call
contig_length
min_score
avg_score
synteny
clade
melded
taxonomy
loci
"""

c_formats["unclassified"] = """
contig_name
call
contig_length
loci
"""

c_formats["details"] = """
contig_name
iteration
clade
gene_scores
gene_spans
"""

# convert to rows
for name, items in c_formats.items( ):
    c_formats[name] = [k for k in items.split( "\n" ) if k != ""]

# subset everything except the details format
c_main_formats = {k:c_formats[k] for k in "lgt no_lgt unclassified".split( )}

# ---------------------------------------------------------------
# cli
# ---------------------------------------------------------------

def get_args( ):

    parser = argparse.ArgumentParser(
        description=description, 
        formatter_class=argparse.RawTextHelpFormatter,
        )

    # input params
    g = parser.add_argument_group( "required inputs" )
    g.add_argument(
        "contigs",
        help="contigs file (.fasta format)",
        )
    g.add_argument(
        "blastout",
        help="output of waafle_search for one set of contigs (.blastout)",
        )
    g.add_argument(
        "gff",
        help="gene calls (from waafle_genecaller or user-supplied) for <contigs> (.gff)",
        )
    g.add_argument(
        "taxonomy",
        help="taxonomy file for the blast database used to make <blastout>",
        )

    # output params
    g = parser.add_argument_group( "output formatting" )
    g.add_argument(
        "--outdir",
        default=".",
        metavar="<path>",
        help="directory for writing output files\n[default: .]",
        )
    g.add_argument(
        "--basename",
        default=None,
        metavar="<str>",
        help="basename for output files\n[default: derived from input]",
        )
    g.add_argument(
        "--write-details",
        action="store_true",
        help="make an additional output file with per-gene clade scores\n[default: off]",
        )
    g.add_argument(
        "--quiet",
        action="store_true",
        help="don't show running progress\n[default: off]",
        )

    # main waafle params
    g = parser.add_argument_group( "main parameters" )
    g.add_argument(
        "-k1", "--one-clade-threshold",
        type=float,
        default=0.5,
        metavar="<0.0-1.0>",
        help="minimum per-gene score for explaining a contig with a single clade\n[default: 0.5]",
        )
    g.add_argument(
        "-k2", "--two-clade-threshold",
        type=float,
        default=0.8,
        metavar="<0.0-1.0>",
        help="minimum per-gene score for explaining a contig with a pair of clades (putative LGT)\n[default: 0.8]",
        )
    g.add_argument(
        "--disambiguate-one",
        choices=["report-best", "meld"],
        default="meld",
        metavar="<report-best/meld>",
        help="what to do when other one-clade explanations fall within <--range> of the best explanation\n[default: meld]",
        )
    g.add_argument(
        "--disambiguate-two",
        choices=["report-best", "jump", "meld"],
        default="meld",
        metavar="<report-best/jump/meld>",
        help="what to do when other two-clade explanations fall within <--range> of the best explanation\n[default: meld]",
        )
    g.add_argument(
        "--range",
        type=float,
        default=0.05,
        metavar="<float>",
        help="when disambiguating, consider explanations within <--range> of the best explanation\n[default: 0.05]",
        )
    g.add_argument(
        "--jump-taxonomy",
        type=int,
        default=None,
        metavar="<1-N>",
        help="before starting, perform 1+ 'jumps' up the taxonomy (e.g. species->genus)\n[default: off]",
        )

    # filters
    g = parser.add_argument_group( "post-detection LGT filters" )
    g.add_argument(
        "--allow-lca",
        action="store_true",
        help="when melding LGT clades, allow the LGT LCA to occur as a melded clade\n[default: off]",
        )
    g.add_argument(
        "--ambiguous-fraction",
        type=float,
        default=0.1,
        metavar="<0.0-1.0>",
        help="allowed fraction of ambiguous (A OR B) gene length in a putative A+B contig\n[default: 0.1]",
        )
    g.add_argument(
        "--ambiguous-threshold",
        choices=["off", "lenient", "strict"],
        default="lenient",
        metavar="<off/lenient/strict>",
        help="homology threshold for defining an ambiguous (A OR B) gene\n[default: lenient]",
        )
    g.add_argument(
        "--sister-penalty",
        choices=["off", "lenient", "strict"],
        default="strict",
        metavar="<off/lenient/strict>",
        help="penalize homologs of missing genes in sisters of LGT clades (or just recipient if known)\n[default: strict]",
        )
    g.add_argument(
        "--clade-genes",
        type=int,
        default=None,
        metavar="<1-N>",
        help="required minimum genes assigned to each LGT clade\n[default: off]",
        )
    g.add_argument(
        "--clade-leaves",
        type=int,
        default=None,
        metavar="<1-N>",
        help="required minimum leaf count supporting each LGT clade (or just recipient if known)\n[default: off]",
        )

    # params related to hit-gene merger
    g = parser.add_argument_group( "gene-hit merge parameters" )
    g.add_argument(
        "--weak-loci",
        choices=["ignore", "penalize", "assign-unknown"],
        default="ignore",
        metavar="<ignore/penalize/assign-unknown>",
        help="method for handling loci that are never assigned to known clades\n[default: ignore]",
        )
    g.add_argument(
        "--annotation-threshold",
        choices=["off", "lenient", "strict"],
        default="lenient",
        metavar="<off/lenient/strict>",
        help="stringency of gene annotation transfer to loci\n[default: lenient]",
        )
    g.add_argument(
        "--min-overlap",
        type=float,
        default=0.1,
        metavar="<0.0-1.0>",
        help="only merge hits into genes if the longer of the two covers this portion of the shorter\n[default: 0.1]",
        )

    # ****some gene-level params imported from waafle_genecaller for consistency****
    attach_shared_args( g )

    # wrap-up
    args = parser.parse_args( )
    return args

# ---------------------------------------------------------------
# object to store contig-specific data
# ---------------------------------------------------------------

class Contig( ):
    
    def __init__( self, name, args ):

        # name of the contig
        self.name = name
        # program arguments
        self.args = args
        # position of the contig in the original file
        self.index = None
        # length of the contig
        self.length = None
        # sequential locus object for each gene on contig
        self.loci = []
        # may be defined later as an array of loci to NOT ignore
        self.mask = None
        # map: locus name -> object
        self.locus_map = {}
        # map: clade -> locus -> np array of per-site scores
        self.site_scores = {}
        # map: clade -> np array of per-gene scores
        self.gene_scores = {}
        # set of clades _currently_ represented on the contig
        self.clades = set( )
        # best one-clade explanation for the contig (Option object)
        self.best_one = None
        # best two-clade explanation for the contig (Option object)
        self.best_two = None
        # homology thresholds (handle corner case of k1>k2)
        self.min_threshold = min( args.one_clade_threshold, args.two_clade_threshold )
        self.max_threshold = max( args.one_clade_threshold, args.two_clade_threshold )
        # threshold for annotation transfer
        if args.annotation_threshold == "off":
            self.annotation_threshold = c_eps
        elif args.annotation_threshold == "lenient":
            self.annotation_threshold = self.min_threshold
        elif args.annotation_threshold == "strict":
            self.annotation_threshold = self.max_threshold

    def attach_loci( self, loci ):
        """ Filter and load genes from GFF into this contig """
        for L in loci:
            if len( L ) >= self.args.min_gene_length:
                self.loci.append( L )
        loci = sorted( loci, key=lambda L: L.start )
        for i, L in enumerate( loci ):
            # assign a name to the locus for convenience
            L.name = str( i + 1 )
            self.locus_map[L.name] = L

    def attach_hits( self, hits ):
        """ Figure out which BLAST hits correspond to which loci """
        for H in hits:
            if H.scov_modified >= self.args.min_scov:
                for i, L in enumerate( self.loci ):
                    # note: <s>strand in Hit specifically
                    if self.args.stranded and H.sstrand != L.strand:
                        continue
                    if hit_locus_overlap( H, L ) >= self.args.min_overlap:
                        self.score_hit( H, L )
        self.clades = set( self.site_scores.keys( ) )
                    
    def score_hit( self, hit, locus ):
        """ Assign site-level scores for hit to appropriate locus """
        l1, l2 = sorted( [locus.start, locus.end] )
        h1, h2 = sorted( [hit.qstart, hit.qend] )
        # get hit into this gene's coordinate system
        h1 = max( 0, h1 - l1 )
        h2 = min( len( locus ) - 1, h2 - l1 )
        # update gene with this hit's score
        ldict = self.site_scores.setdefault( hit.taxon, {} )
        if locus.name not in ldict:
            ldict[locus.name] = np.zeros( len( locus ) )
        ldict[locus.name][h1:h2+1] = np.maximum( ldict[locus.name][h1:h2+1], hit.waafle_score )
        # (potentially) update locus with this hit's annotations
        for system, value in hit.annotations.items( ):
            ref = locus.annotation_scores.get( system, self.annotation_threshold )
            if ref is None:
                # annotation from GFF, don't overwrite
                continue
            elif hit.waafle_score >= ref:
                # add/replace
                locus.annotations[system] = value
                locus.annotation_scores[system] = hit.waafle_score

    def update_gene_scores( self ):
        """ Convert nucleotide-level clade scores to gene scores, handle bad loci """
        # clear existing scores (if any)
        self.gene_scores = {}
        # collapse site scores to genes by averaging
        for clade, ldict in self.site_scores.items( ):
            scores = []
            for locus in self.loci:
                if locus.name in ldict:
                    scores.append( np.mean( ldict[locus.name] ) )
                else:
                    scores.append( 0 )
            self.gene_scores[clade] = np.array( scores )
        # compute max per-gene scores (used below)
        maxes = np.zeros( len( self.loci ) )
        for clade, values in self.gene_scores.items( ):
            if clade != wu.c_unknown:
                maxes = np.maximum( maxes, values )
        # weak loci option 1: do nothing
        if self.args.weak_loci == "penalize":
            pass
        # weak loci option 2: spike unknown taxon
        elif self.args.weak_loci == "assign-unknown":
            self.gene_scores[wu.c_unknown] = 1 - maxes
            self.clades.add( wu.c_unknown )
        # weak loci option 3: mask
        elif self.args.weak_loci == "ignore":
            ok_list = []
            for index, value in enumerate( maxes ):
                self.loci[index].ignore = True
                if value >= self.min_threshold:
                    ok_list.append( index )
                    self.loci[index].ignore = False
            self.mask = None if len( ok_list ) == len( self.loci ) else np.array( ok_list )
        # update contig clades (will include "Unknown" if added above)
        self.clades = {clade for clade in self.gene_scores}

    def raise_taxonomy( self, taxonomy ):
        """ Recombines hits according to clade parents, then converts to new gene scores """
        new_site_scores = {}
        for clade, ldict in self.site_scores.items( ):
            parent = taxonomy.get_parent( clade )
            if parent is None:
                continue
            inner = new_site_scores.setdefault( parent, {} )
            for name in ldict:
                if name not in inner:
                    inner[name] = np.zeros( len( self.locus_map[name] ) )
                inner[name] = np.maximum( inner[name], ldict[name] )
        self.site_scores = new_site_scores
        # use the new site scores to update gene scores
        self.update_gene_scores( )

    def score( self, clade1, clade2=None ):
        """ Score one or two clades over the contig """
        # crit is the MIN max score at each locus over the clade(s)
        crit = None
        # rank is the AVG max score at each locus over the clade(s)
        rank = None
        maxes = scores1 = self.gene_scores[clade1]
        if clade2 is not None:
            scores2 = self.gene_scores[clade2]
            maxes = np.maximum( scores1, scores2 )
        # ignore masked positions, if applicable
        maxes = maxes if self.mask is None else maxes[self.mask]
        crit = np.min( maxes )
        rank = np.mean( maxes )
        return crit, rank

# ---------------------------------------------------------------
# contig explanation object
# ---------------------------------------------------------------
            
class Option( ):

    def __init__( self, contig ):
        # link to the associated contig
        self.contig    = contig
        # False if a contig has failed 1+ LGT filters
        self.ok        = True
        # the one- or two-clade critical value
        self.crit      = None
        # the one- or two-clade rank
        self.rank      = None
        # 1st clade (A)
        self.clade1    = None
        # 2nd clade (B)
        self.clade2    = None
        # synteny pattern (e.g. AABBA)
        self.synteny   = None
        # direction of transfer, if known ("B>A")
        self.direction = "A?B"
        # donor if known (B in ^A+B+A+$)
        self.donor     = None
        # recip if known (A in ^A+B+A+$)
        self.recip     = None
        # clade1 tails after melding (e.g. species melded into genus)
        self.tails1    = []
        # clade2 tails after melding
        self.tails2    = []

    def set_synteny_one( self ):
        # load k params from contig / args
        k1 = self.contig.args.one_clade_threshold
        # clade1 gene scores
        scores = self.contig.gene_scores[self.clade1]
        # generate synteny
        synteny = ""
        for s, L in zip( scores, self.contig.loci ):
            if L.ignore:
                synteny += c_synchar_ignored
            elif s >= k1:
                synteny += "A"
            else:
                synteny += c_synchar_error
        self.synteny = synteny

    def set_synteny_two( self ):
        # load k params from contig / args
        k1 = self.contig.args.one_clade_threshold
        k2 = self.contig.args.two_clade_threshold
        switch = {"off":c_eps, "lenient":min( k1, k2 ), "strict":max( k1, k2 )}
        k_ambiguous = switch[self.contig.args.ambiguous_threshold]
        # clade1/2 gene scores
        scores1 = self.contig.gene_scores[self.clade1]
        scores2 = self.contig.gene_scores[self.clade2]
        # don't consider "ambiguous" option if unknown taxon involved
        unknown_involved = wu.c_unknown in [self.clade1, self.clade2]
        # generate synteny
        synteny = ""
        for s1, s2, L in zip( scores1, scores2, self.contig.loci ):
            if L.ignore:
                synteny += c_synchar_ignored
            elif min( s1, s2 ) >= k_ambiguous and not unknown_involved:
                synteny += c_synchar_ambiguous
            elif s1 >= k2:
                synteny += "A"
            elif s2 >= k2:
                synteny += "B"
            else:
                synteny += c_synchar_error
        self.synteny = synteny
        # force "A" to be the clade at the first locus of clear taxonomy
        if re.search( "^[^A]*B", self.synteny ):
            self.clade1, self.clade2 = self.clade2, self.clade1
            switch = {"A":"B", "B":"A"}
            self.synteny = "".join( [switch.get( char, char ) for char in self.synteny] )
        # direction clear? (don't penalize 'ignored' genes)
        if re.search( "^A+B+A+$", self.synteny.replace( c_synchar_ignored, "" ) ):
            self.direction = "B>A"
            self.donor = self.clade1
            self.recip = self.clade2

# ---------------------------------------------------------------
# helper functions
# ---------------------------------------------------------------

def is_ok( option ):
    """ treat None option as not OK """
    return (option is not None and option.ok)

def not_ok( option ):
    """ treat None option as not OK """
    return (option is None or not option.ok)

def hit_locus_overlap( hit, locus ):
    # query = contig; same coordinate space as locus
    a1, a2 = hit.qstart, hit.qend
    b1, b2 = locus.start, locus.end
    # calc_overlap sorts start/end internally
    return wu.calc_overlap( a1, a2, b1, b2 )

def evaluate_contig( contig, taxonomy, details, args ):
    iteration = 1
    write_details( details, contig, iteration )
    best_one = explain_one( contig, taxonomy, args )
    best_two = explain_two( contig, taxonomy, args ) if not_ok( best_one ) else None
    while len( contig.clades ) > 0 and \
            "r__Root" not in contig.clades and \
            not_ok( best_one ) and \
            not_ok( best_two ):
        contig.raise_taxonomy( taxonomy )
        write_details( details, contig, iteration )
        best_one = explain_one( contig, taxonomy, args )
        best_two = explain_two( contig, taxonomy, args ) if not_ok( best_one ) else None
        iteration += 1
        if iteration > 100:
            wu.die( "  Warning: Runaway taxonomic recursion for", contig.name )
    contig.best_one = best_one
    contig.best_two = best_two

def explain_one( contig, taxonomy, args ):
    options = []
    for clade in contig.clades:
        crit, rank = contig.score( clade )
        if crit >= args.one_clade_threshold:
            option = Option( contig )
            option.crit    = crit
            option.rank    = rank
            option.clade1  = clade
            option.set_synteny_one( )
            options.append( option )
    best = meld_one( options, taxonomy, args ) if len( options ) > 0 else None
    return best

def explain_two( contig, taxonomy, args ):
    options = []
    potential_clades = []
    # speedup: skip clades that never exceed k2 threshold
    for clade in contig.clades:
        if max( contig.gene_scores[clade] ) >= args.two_clade_threshold:
            potential_clades.append( clade )
    for clade1 in potential_clades:
        for clade2 in potential_clades:
            if clade1 < clade2:
                crit, rank = contig.score( clade1, clade2 )
                if crit >= args.two_clade_threshold:
                    option = Option( contig )
                    option.rank    = rank
                    option.crit    = crit
                    option.clade1  = clade1
                    option.clade2  = clade2
                    option.set_synteny_two( )
                    options.append( option )
    best = meld_two( options, taxonomy, args ) if len( options ) > 0 else None
    return best

def meld_one( options, taxonomy, args ):
    # note: unlike LGTs, "melds" never invalidate a one-bug explanation
    options = sorted( options, key=lambda x: x.rank )
    best = options[-1]
    options = [k for k in options if best.rank - k.rank <= args.range]
    # meld other options into best?
    if args.disambiguate_one == "meld":
        to_meld = [k.clade1 for k in options]
        best.clade1 = taxonomy.get_lca( *to_meld )
        best.tails1 = taxonomy.get_tails( to_meld, best.clade1 )
    return best

def meld_two( options, taxonomy, args ):
    options = sorted( options, key=lambda o: o.rank )
    best = options[-1]
    options = [k for k in options if best.rank - k.rank <= args.range]
    # apply lgt filters to options individually
    for o in options:
        apply_lgt_checks( o, taxonomy, args )
    # if only one option, no disambiguation needed
    if len( options ) == 1:
        best = best
    # multiple options + best: keep best
    elif args.disambiguate_two == "report-best":
        best = best
    # multiple options + jump: kill best
    elif args.disambiguate_two == "jump":
        best = None
    # multiple options + meld, but inconsistent options: kill best
    elif args.disambiguate_two == "meld" and not meld_precheck( options ):
        best = None
    # multiple options + meld: merge options into best
    elif args.disambiguate_two == "meld":
        clades1 = [o.clade1 for o in options]
        clades2 = [o.clade2 for o in options]
        best.clade1 = lca1 = taxonomy.get_lca( *clades1 )
        best.clade2 = lca2 = taxonomy.get_lca( *clades2 )
        best.tails1 = taxonomy.get_tails( clades1, lca1 )
        best.tails2 = taxonomy.get_tails( clades2, lca2 )
        # post-meld lca check
        if not args.allow_lca:
            new_clades = [best.clade1, best.clade2]
            new_lca = taxonomy.get_lca( *new_clades )
            if new_lca in new_clades:
                best = None
    # should never happen...
    else:
        wu.die( "Unexpected two-clade resolution." )
    return best

def meld_precheck( options ):
    ret = True
    for o in options:
        if not_ok( o ) or o.synteny != options[0].synteny:
            ret = False
    return ret

def apply_lgt_checks( option, taxonomy, args ):
    if args.ambiguous_fraction is not None:
        check_ambiguous_fraction( option, args )
    if args.clade_genes is not None:
        check_clade_genes( option, args )
    if args.clade_leaves is not None:
        check_clade_leaves( option, taxonomy, args )
    # note: different trigger
    if args.sister_penalty != "off":
        check_sister_penalty( option, taxonomy, args )

# ---------------------------------------------------------------
# LGT filtering functions
# ---------------------------------------------------------------

def check_ambiguous_fraction( option, args ):
    total_len = 0
    ambiguous_len = 0
    for char, locus in zip( option.synteny, option.contig.loci ): 
        if char in ["A", "B", c_synchar_ambiguous]:
            total_len += len( locus )
            ambiguous_len += len( locus ) if char == c_synchar_ambiguous else 0
    test = ambiguous_len / float( total_len )
    if test > args.ambiguous_fraction:
        option.ok = False

def check_clade_genes( option, args ):
    clade_genes = Counter( option.synteny )
    test = min( [clade_genes.get( char, 0 ) for char in "AB"] )
    if test < args.clade_genes:
        option.ok = False

def check_clade_leaves( option, taxonomy, args ):
    # only apply to recipient if known
    to_check = [option.recip] if option.recip is not None else [option.clade1, option.clade2]
    test = min( [taxonomy.get_leaf_count( clade ) for clade in to_check] )
    if test < args.clade_leaves:
        option.ok = False

def check_sister_penalty( option, taxonomy, args ):
    C = option.contig
    # strict punishes sister homologs >k1; lenient punishes >k2 (happens less often)
    switch = {"lenient":C.max_threshold, "strict":C.min_threshold}
    my_threshold = switch[args.sister_penalty]
    clade1, clade2 = option.clade1, option.clade2
    sisters, penalties = {}, {}
    # note the unintuitive swap: a B locus is penalized by A's sisters (but not A itself)
    sisters["B"] = taxonomy.get_sisters( clade1 ) - {clade2}
    sisters["A"] = taxonomy.get_sisters( clade2 ) - {clade1}
    for i, char in enumerate( option.synteny ):
        if char not in sisters:
            continue
        hits = 0
        for clade in sisters[char]:
            if clade in C.gene_scores:
                if C.gene_scores[clade][i] >= my_threshold:
                    hits += 1
        if len( sisters[char] ) > 0:
            hits /= float( len( sisters[char] ) )
        penalties.setdefault( char, [] ).append( hits )
    # mean prevalence of sister homologs; we will count >0 as bad
    penalties = {char:np.mean( values ) for char, values in penalties.items( )}
    # only penalize donor (B) loci if recipient known
    to_check = "B" if option.recip is not None else "AB"
    test = max( [penalties.get( char, 0 ) for char in to_check] )
    if test > 0:
        option.ok = False

# ---------------------------------------------------------------
# output formatting
# ---------------------------------------------------------------

def make_tails_field( tails ):
    """ make string representation of the melded clades """
    ret = ""
    if tails is not None:
        items = set( )
        for t in tails:
            if len( t ) > 0:
                items.add( c_delim2.join( t ) )
        ret = c_delim1.join( sorted( items ) )
    return ret

def make_loci_field( loci ):
    """ make string representation of contig loci """
    codes = [L.code for L in loci]
    return c_delim2.join( codes )

def make_gene_scores_field( contig, clade ):
    scores = ["{A:.{B}f}".format( A=s, B=c_precision ) \
                  for s in contig.gene_scores[clade]]
    return c_delim2.join( scores )

def make_gene_spans_field( contig, clade ):
    gene_spans = []
    for L in contig.loci:
        if clade not in contig.site_scores:
            # no site scores for this clade, e.g. spiked unknown
            gene_spans.append( c_missing_annotation )
        elif L.name not in contig.site_scores[clade]:
            # no hits for this gene in this clade
            gene_spans.append( c_missing_annotation )
        else:
            # scores exist
            site_scores = contig.site_scores[clade][L.name]
            # non-zero indices (base-1)
            nzi = 1 + np.nonzero( site_scores )[0]
            # isolate interesting indices (step-1 to left or right but not both)
            L_diff = np.concatenate( [[0], nzi[1:] - nzi[0:-1]] )
            R_diff = np.concatenate( [nzi[1:] - nzi[0:-1], [0]] )
            select = np.logical_xor( L_diff==1, R_diff==1 )
            nzi = nzi[select]
            gene_spans.append( c_delim3.join( [str( k ) for k in nzi] ) )
    return c_delim2.join( gene_spans )

def attach_rowdict_functions( rowdict, contig, systems ):
    """ update rowdict with contig functions """
    for s in systems:
        items = []
        for locus in contig.loci:
            items.append( locus.annotations.get( s, c_missing_annotation ) )
        # note: key here must match definition in headers
        rowdict[c_annotation_prefix + s] = c_delim2.join( items )

def write_details( details, contig, iteration ):
    if details is not None:
        for clade in contig.clades:
            rowdict = {
                "contig_name": contig.name,
                "iteration":   iteration,
                "clade":       clade,
                "gene_scores": make_gene_scores_field( contig, clade ),
                "gene_spans":  make_gene_spans_field( contig, clade ),
                }
            wu.write_rowdict( rowdict, c_formats["details"], details )

def write_main_output_files( contigs, taxonomy, args ):

    # open output file handles
    wu.say( "Initializing outputs." )
    if args.basename is None:
        args.basename = os.path.split( args.blastout )[1].split( "." )[0]
    handles = {}
    for option in ["lgt", "no_lgt", "unclassified"]:
        file_name = ".".join( [args.basename, option, "tsv"] )
        handles[option] = open( os.path.join( args.outdir, file_name ), "w" )

    # determine possible function annotation systems
    systems = set( )
    for contig in contigs.values( ):
        for locus in contig.loci:
            for system in locus.annotations:
                systems.add( system )
    for option in c_main_formats:
        for s in sorted( systems ):
            c_formats[option].append( c_annotation_prefix + s )

    # print headers
    for name in handles:
        wu.write_rowdict( None, c_main_formats[name], file=handles[name] )
        
    # write results (sorted loop over contigs)
    for contig_name in sorted( contigs ):
        contig = contigs[contig_name]
        best_one = contig.best_one
        best_two = contig.best_two
        # unclassified
        if not_ok( best_one ) and not_ok( best_two ):
            rowdict = {
                "contig_name":   contig_name,
                "call":          "unclassified",
                "contig_length": contig.length,
                "loci":          make_loci_field( contig.loci ),
                }
            attach_rowdict_functions( rowdict, contig, systems )
            wu.write_rowdict( rowdict, c_formats["unclassified"], handles["unclassified"] )
        # no_lgt
        elif is_ok( best_one ):
            clade = best_one.clade1
            rowdict = {
                "contig_name":   contig_name,
                "call":          "no_lgt",
                "contig_length": contig.length,
                "min_score":     best_one.crit,
                "avg_score":     best_one.rank,
                "synteny":       best_one.synteny,
                "clade":         clade,
                "taxonomy":      c_delim2.join( taxonomy.get_lineage( clade ) ),
                "melded":        make_tails_field( best_one.tails1 ),
                "loci":          make_loci_field( contig.loci ),
                }
            attach_rowdict_functions( rowdict, contig, systems )
            wu.write_rowdict( rowdict, c_formats["no_lgt"], handles["no_lgt"] )
        # lgt
        elif is_ok( best_two ):
            clade1, clade2 = best_two.clade1, best_two.clade2
            rowdict = {
                "contig_name":   contig_name,
                "call":         "lgt",
                "contig_length": contig.length,
                "min_max_score": best_two.crit,
                "avg_max_score": best_two.rank,
                "synteny":       best_two.synteny,
                "direction":     best_two.direction,
                "clade_A":       clade1, 
                "clade_B":       clade2,
                "lca":           taxonomy.get_lca( clade1, clade2 ),
                "taxonomy_A":    c_delim2.join( taxonomy.get_lineage( clade1 ) ),
                "taxonomy_B":    c_delim2.join( taxonomy.get_lineage( clade2 ) ),
                "melded_A":      make_tails_field( best_two.tails1 ),
                "melded_B":      make_tails_field( best_two.tails2 ),
                "loci":          make_loci_field( contig.loci ),
                }
            attach_rowdict_functions( rowdict, contig, systems )
            wu.write_rowdict( rowdict, c_formats["lgt"], handles["lgt"] )

    # wrap up
    for h in handles.values( ):
        h.close( )

# ---------------------------------------------------------------
# main
# ---------------------------------------------------------------

def main( ):

    args = get_args( )
    wu.say( "Loading taxonomy." )
    taxonomy = wu.Taxonomy( args.taxonomy )

    # initialize contigs
    wu.say( "Initializing contigs." )
    contigs = {}
    contig_lengths = wu.read_contig_lengths( args.contigs )
    index = 0
    for contig_name, length in contig_lengths.items( ):
        C = Contig( contig_name, args )
        C.length = length
        index += 1
        C.index = index
        contigs[contig_name] = C

    # process gff
    wu.say( "Adding gene coordinates." )
    for contig_name, loci in wu.iter_contig_loci( args.gff, attach_annotations=False ):
        if contig_name not in contigs:
            wu.say( "  Unknown contig in <gff> file", contig_name )
            continue
        C = contigs[contig_name]
        C.attach_loci( loci )

    # prepare details file
    details = None
    if args.write_details:
        details = wu.try_open( os.path.join( 
                args.outdir, args.basename + ".details.tsv.gz" ), "w" )
        # headers
        wu.write_rowdict( None, c_formats["details"], file=details )

    # parse hits, process contigs
    wu.say( "Analyzing contigs." )

    # major contig loop
    for contig_name, hits in wu.iter_contig_hits( args.blastout ):
        if contig_name not in contigs:
            wu.say( "  Unknown contig in <blastout> file", contig_name )
            continue
        # this is a good contig
        C = contigs[contig_name]
        if not args.quiet:
            wu.say( "  #{:>7,} of {:>7,}".format( C.index, len( contigs ) ) )
        # attach hits to genes
        C.attach_hits( hits )
        C.update_gene_scores( )
        # initial jumps?
        if args.jump_taxonomy is not None:
            for j in range( args.jump_taxonomy ):
                C.raise_taxonomy( taxonomy )
        # evaluate; note: the 'ignore' option can result in "empty" contigs
        if not all( [L.ignore for L in C.loci] ):
            evaluate_contig( C, taxonomy, details, args )

    # wrap up
    write_main_output_files( contigs, taxonomy, args )
    wu.say( "Finished successfully." )
    if details is not None:
        details.close( )
                    
if __name__ == "__main__":
    main( )



#!/usr/bin/env python

"""
This module is a part of:
WAAFLE, a [W]orkflow to [A]nnotate [A]ssemblies and [F]ind [L]GT [E]vents

Copyright (c) 2019 Harvard T.H. Chan School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function # Python 2.7+ required
import os
import sys
import csv
import argparse

from waafle import utils as wu

# ---------------------------------------------------------------
# description
# ---------------------------------------------------------------

description = wu.describe( """
{SCRIPT}: (Optional) Step 1.5 in the WAAFLE pipeline

Use the results of waafle_search to identify candidate gene
loci in a set of contigs and output them as a GFF file for use
in the next step. Users can optionally supply their own (independently-generated)
GFF file.
""" )

# ---------------------------------------------------------------
# cli
# ---------------------------------------------------------------

def get_args( ):
    """ Get arguments passed to script """
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument( 
        "blastout",
        help="(custom) blast output from waafle_search",
        )
    parser.add_argument( 
        "--gff",
        default=None,
        metavar="<path>",
        help="path for (output) waafle gene calls (.gff)\n[default: <derived from input>]",
        )
    parser.add_argument(
        "--min-overlap",
        default=0.1,
        type=float,
        metavar="<float>",
        help=("if a large hit covers this fraction of a smaller hit, "
              "consider them part of the same gene group\n[default: 0.1]"),
        )
    attach_shared_args( parser )
    args = parser.parse_args( )
    return args

def attach_shared_args( parser ):
    """ these arguments are shared with orgscorer """
    parser.add_argument(
        "--min-gene-length",
        default=200,
        type=float,
        metavar="<int>",
        help="minimum allowed gene length\n[default: 200]",
        )
    parser.add_argument(
        "--min-scov",
        default=0.75,
        type=float,
        metavar="<float>",
        help="(modified) scoverage filter for hits to gene catalog\n[default: 0.75]",
        )
    parser.add_argument(
        "--stranded",
        action="store_true",
        help="only merge hits into hits/genes of the same strandedness\n[default: off]",
        )

# ---------------------------------------------------------------
# utilities
# ---------------------------------------------------------------

def hits2ints( hits, scov ):
    """ filter hits by scoverage; convert to intervals """
    intervals = []
    for hit in hits:
        if hit.scov_modified >= scov:
            intervals.append( [hit.qstart, hit.qend, hit.sstrand] )
    return intervals

def overlap_inodes( inode1, inode2 ):
    """ compute overlap between two intervals """
    a1, b1 = inode1.start, inode1.stop
    a2, b2 = inode2.start, inode2.stop
    overlap = wu.calc_overlap( a1, b1, a2, b2 )
    return overlap
    
def merge_inodes( *inodes ):
    """ merge overlapping intervals into a single node """
    start = stop = None
    strand_rank = []
    for inode in inodes:
        if start is None or inode.start < start:
            start = inode.start
        if stop is None or inode.stop > stop:
            stop = inode.stop
        strand_rank.append( [len( inode ), inode.strand] )
    # assign strand of largest interval to the merged interval
    strand = sorted( strand_rank )[-1][1]
    return wu.INode( start, stop, strand )

def make_inodes( intervals ):
    """ convert list of intervals to list of inodes """
    inodes = []
    for interval in intervals:
        start, stop = interval[0:2]
        strand = "+" if len( interval ) < 3 else interval[2]
        inodes.append( wu.INode( start, stop, strand ) )
    return inodes
        
def overlap_intervals( intervals, threshold, stranded ):
    """ find and collapse overlapping intervals """
    inodes = make_inodes( intervals )
    inodes = sorted( inodes, key=lambda inode: inode.start )
    for index, inode1 in enumerate( inodes ):
        for inode2 in inodes[index+1:]:
            if not stranded or inode1.strand == inode2.strand:
                score = overlap_inodes( inode1, inode2 )
                if score >= threshold:
                    # store as edge in network
                    inode1.attach( inode2 )
                    inode2.attach( inode1 )
                elif score == 0:
                    # no further inode2 can overlap this inode1
                    break
    # divide intervals into "connected components"
    ccs = []
    for inode in inodes:
        if not inode.visited:
            ccs.append( inode.get_connected_component( ) )
    # merge intervals and report as simple lists
    new_intervals = []
    for cc in ccs:
        merged = merge_inodes( *cc ).to_list( )
        new_intervals.append( merged )
    return new_intervals

# ---------------------------------------------------------------
# main
# ---------------------------------------------------------------

"""
NOTES:

1) Organize BLAST hits that pass the scov filter into intervals
2) Sort intervals by start site and find connected (overlapping) intervals
3) Merge connected intervals into genes, and filter by gene length
4) Print genes as GFF

GFF FORMAT:

1. seqname
2. source
3. feature
4. start
5. end
6. score
7. strand
9. frame
8. attribute
"""

def main( ):

    args = get_args( )
    if args.gff is None:
        name = wu.path2name( args.blastout )
        args.gff = wu.name2path( name, ".", ".gff" )

    fh_gff = wu.try_open( args.gff, "w" )
    writer = csv.writer( fh_gff, csv.excel_tab )

    for contig, hits in wu.iter_contig_hits( args.blastout ):
        intervals = hits2ints( 
            hits, 
            args.min_scov,
            )
        intervals = overlap_intervals( 
            intervals, 
            args.min_overlap, 
            args.stranded == "on",
            )
        for start, stop, strand in intervals:
            gene_length = stop - start + 1
            if gene_length >= args.min_gene_length:
                items = [
                    contig,
                    "waafle_genecaller",
                    "gene",
                    start,
                    stop,
                    ".",
                    strand,
                    0,
                    ".",
                    ]
                writer.writerow( [str( k ) for k in items] )

    fh_gff.close( )
    wu.say( "Finished successfully." )

if __name__ == "__main__":
    main( )



#!/usr/bin/env python

"""
This module is a part of:
WAAFLE, a [W]orkflow to [A]nnotate [A]ssemblies and [F]ind [L]GT [E]vents

Copyright (c) 2019 Harvard T.H. Chan School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function # Python 2.7+ required

import os
import sys
import csv
import re
import gzip
import bz2
from collections import OrderedDict

import numpy as np

# ---------------------------------------------------------------
# ---------------------------------------------------------------
# GENERIC HELPER CLASSES / FUNCTIONS
# ---------------------------------------------------------------
# ---------------------------------------------------------------

def say( *args ):
    print( " ".join( map( str, args ) ), file=sys.stderr )

def die( *args ):
    args = ["LETHAL ERROR:"] + list( args )
    say( *args )
    sys.exit( "EXITING." )

def path2name( p ):
    return os.path.split( p )[1].split( "." )[0]

def name2path( name, root=".", ext="" ):
    return os.path.join( root, name+ext )

def try_open( path, *args ):
    """ Open a file; fail gracefully """
    fh = None
    try:
        if path.endswith( ".gz" ):
            fh = gzip.GzipFile( path, *args )
        elif path.endswith( ".bz2" ):
            fh = bz2.BZ2File( path, *args )
        else:
            fh = open( path, *args )
    except:
        sys.exit( "Can't open file: {}".format( path ) )
    return fh

def describe( text, width=80, margin=2 ):
    margin = " " * margin
    # remove flanking whitespace
    text = text.strip( )
    # insert script name
    text = text.format( SCRIPT=os.path.split( sys.argv[0] )[1] )
    lines = text.split( "\n" )
    # title
    rule = "=" * width
    newlines = [rule, margin + lines[0], rule, "\n"]
    newline = margin
    # description lines
    for line in lines[2:]:
        line = line.strip( )
        if line == "":
            newlines.append( newline )
            newlines.append( "\n" )
            newline = margin
            continue
        words = line.split( )
        for word in words:
            if len( word ) > width:
                newlines.append( newline )
                newlines.append( word )
                newline = margin
            elif len( newline + " " + word ) > width:
                newlines.append( newline )
                newline = margin + word
            else:
                newline += (" " if newline != margin else "") + word
    if len( newline ) > 0:
        newlines.append( newline )
    newlines += ["\n", rule]
    return "\n".join( [k for k in newlines] )

def read_contig_lengths( fasta ):
    data = OrderedDict( )
    header = None
    with try_open( fasta ) as fh:
        for line in fh:
            line = line.strip( )
            if line[0] == ">":
                header = line[1:].split( )[0]
                data[header] = 0
            else:
                data[header] += len( line )
    return data

def write_rowdict( rowdict=None, format=None, file=None, 
                   delim="\t", precision=4, empty_field="--", ):
    """ write rowdict line conditioned on format """
    if rowdict is None:
        # print headers
        print( delim.join( [k.upper( ) for k in format] ), file=file )
    elif set( rowdict ) != set( format ):
        # bad line
        for f in format:
            print( f, rowdict.get( f, None ) )
        die( "Format mismatch." )
    else:
        # print line
        items = []
        for f in format:
            if type( rowdict[f] ) in [float, np.float32, np.float64]:
                rowdict[f] = "{A:.{B}f}".format( A=rowdict[f], B=precision )
            items.append( str( rowdict[f] ) if rowdict[f] != "" else empty_field )
        try:
            print( delim.join( items ), file=file )
        except:
            die( "Writing row failed." )

class Frame( ):

    """ manipulate generic tabular data with headers """
    
    def __init__( self, path ):
        self.reader = csv.reader( try_open( path ), dialect="excel-tab" )
        self.headers = next( self.reader )
        
    def iter_rowdicts( self ):
        for row in self.reader:
            yield {h:v for h, v in zip( self.headers, row )}
        
# ---------------------------------------------------------------
# ---------------------------------------------------------------
# WORKING WITH BLAST
# ---------------------------------------------------------------
# ---------------------------------------------------------------

# ---------------------------------------------------------------
# blast constants
# ---------------------------------------------------------------

c_blast_fields = [
    ["qseqid"   , str],
    ["sseqid"   , str],
    ["qlen"     , int],
    ["slen"     , int],
    ["length"   , int],
    ["qstart"   , int],
    ["qend"     , int],
    ["sstart"   , int],
    ["send"     , int],
    ["pident"   , float],
    ["positive" , int],
    ["gaps"     , int],
    ["evalue"   , float],
    ["bitscore" , float],
    ["sstrand"  , str],
]
c_blast_format_string = " ".join( ["6"] + [fname for [fname, ftype] in c_blast_fields] )
# blast default is 500, which is sometimes too small for long contigs
c_max_target_seqs = 10000

# ---------------------------------------------------------------
# blast hit object
# ---------------------------------------------------------------

class Hit( ):

    """
    Processes the information from a single BLAST row.
    Row is provided already split by the csv reader.
    Hit MUST be compatible with the blast format defined above.

    Additional information is pulled into the hit object by parsing the
    subject sequence identifier. The expected format is:
    
    0 : unique gene id
    1 : taxon name
    2+: one or more functional annotations in "system=id" format
    """

    def __init__( self, blast_row ):
        if len( blast_row ) != len( c_blast_fields ):
            die( "inconsistent blast row: {}".format( str( blast_row ) ) )
        # pull values from blast line and coerce to appropriate types
        for [fname, ftype], value in zip( c_blast_fields, blast_row ):
            setattr( self, fname, ftype( value ) )
        # overrides
        self.sstrand = "-" if self.sstrand == "minus" else "+"
        # derived coverage stats
        self.scov = ( abs( self.send - self.sstart ) + 1 ) / float( self.slen )
        self.qcov = ( abs( self.qend - self.qstart ) + 1 ) / float( self.qlen )
        # special scoverage that doesn't penalize hanging off contig end
        if self.sstrand == "-":
            sstart = self.slen - self.sstart + 1
            send   = self.slen - self.send + 1
        else:
            sstart = self.sstart
            send   = self.send
        self.ltrim = max( 0, sstart - self.qstart )
        self.rtrim = max( 0, self.slen - sstart - self.qlen + self.qstart )
        self.scov_modified = (send - sstart + 1) / float( self.slen - self.ltrim - self.rtrim )
        # waafle score
        self.waafle_score = self.scov_modified * self.pident / 100.0 
        # values extracted from subject header
        items = self.sseqid.split( "|" )
        if len( items ) < 2:
            die( "bad subject id header:", self.sseqid )
        self.geneid = items[0]
        self.taxon = items[1]
        # process possible annotations
        self.annotations = {}
        if len( items ) > 2:
            for k in items[2:]:
                system, name = k.split( "=" )
                self.annotations[system] = name

# ---------------------------------------------------------------
# blast utils
# ---------------------------------------------------------------

def iter_hits( blastoutfile ):
    """
    Iterate through the hits in a blast file
    """
    with try_open( blastoutfile ) as fh:                 
        for row in csv.reader( fh, dialect="excel-tab" ):
            yield Hit( row )

def iter_contig_hits( blastoutfile ):
    """
    Iterate through hits by contig (assumes file is sorted by query)
    """
    contig, hits = None, []
    with try_open( blastoutfile ) as fh:                 
        for row in csv.reader( fh, dialect="excel-tab" ):
            hit = Hit( row )
            if contig is not None and hit.qseqid != contig:
                yield contig, hits
                # reset
                hits = []
            contig = hit.qseqid
            hits.append( hit )
        # last case cleanup
        yield contig, hits

# ---------------------------------------------------------------
# ---------------------------------------------------------------
# WORKING WITH GFFS
# ---------------------------------------------------------------
# ---------------------------------------------------------------

# ---------------------------------------------------------------
# constants
# ---------------------------------------------------------------

c_gff_fields = [
    ["seqname"   , str],
    ["source"    , str],
    ["feature"   , str],
    ["start"     , int],
    ["end"       , int],
    ["score"     , float],
    ["strand"    , str],
    ["frame"     , str],
    ["attribute" , str],
]

# ---------------------------------------------------------------
# gff line object (locus)
# ---------------------------------------------------------------

class Locus( ):

    def __init__( self, gff_row, attach_annotations=True ):
        # gff fields
        if len( gff_row ) != len( c_gff_fields ):
            die( "Bad GFF row:", gff_row )
        for [fname, ftype], value in zip( c_gff_fields, gff_row ):
       	    setattr( self, fname, ftype( value ) if value != "." else value )
        # annotations
        self.annotations = {}
        self.annotation_scores = {}
        for item in self.attribute.split( "; " ):
            match = re.search( "^(.*?) \"(.*)\"$", item )
            if attach_annotations and match:
                system, value = match.groups( )
                self.annotations[system] = value
                self.annotation_scores[system] = None
        # no name by default
        self.name = None
        self.code = ":".join( [str( self.start ), str( self.end ), self.strand] )
        # ignore status
        self.ignore = False

    def __len__( self ):
        return abs( self.end - self.start ) + 1

# ---------------------------------------------------------------
# gff utils
# ---------------------------------------------------------------

"""
Note: Currently we'll ignore any annotations in a user-supplied
GFF and focus on those assigned by WAAFLE.
Could update this later.
"""

def iter_loci( gff_file, attach_annotations=True ):
    with try_open( gff_file ) as fh:
        for row in csv.reader( fh, csv.excel_tab ):
            if row[0][0] == "#":
                continue
            yield Locus( row, attach_annotations=attatch_annotations )

def iter_contig_loci( gff_file, attach_annotations=True ):
    contig, loci = None, []
    with try_open( gff_file ) as fh:
        for row in csv.reader( fh, csv.excel_tab ):
            if row[0][0] == "#":
                continue
            l = Locus( row, attach_annotations=attach_annotations )
            if contig is not None and l.seqname != contig:
                yield contig, loci
                # reset
                loci = []
            contig = l.seqname
            loci.append( l )
        # last case cleanup
        yield contig, loci

# ---------------------------------------------------------------
# ---------------------------------------------------------------
# WORKING WITH TAXONOMY
# ---------------------------------------------------------------
# ---------------------------------------------------------------

# ---------------------------------------------------------------
# constants
# ---------------------------------------------------------------

c_unknown = "Unknown"
c_root    = "r__Root"

# ---------------------------------------------------------------
# taxonomy object
# ---------------------------------------------------------------

class Taxonomy( ):

    def __init__( self, path ):
        self.parents = {}
        self.children = {}
        with try_open( path ) as fh:
            for clade, parent in csv.reader( fh, csv.excel_tab ):
                self.parents[clade] = parent
                self.children.setdefault( parent, set( ) ).add( clade )
        # memoizer for get_leaf_count( )
        self.known_leaf_count = {}

    def get_parent( self, clade ):
        return self.parents.get( clade, c_root )

    def get_children( self, clade ):
        return self.children.get( clade, set( ) )

    def get_lineage( self, clade ):
        l = [clade]
        while l[-1] != c_root:
            parent = self.get_parent( clade )
            l.append( parent )
            clade = parent
        l.reverse( )
        return l

    def get_lca( self, *clades ):
        lca = c_root
        lineages = [self.get_lineage( c ) for c in clades]
        min_depth = min( [len( l ) for l in lineages] )
        for i in range( 0, min_depth ):
            level = {l[i] for l in lineages}
            if len( level ) == 1:
                lca = list( level )[0]
            else:
                break
        return lca

    def get_tails( self, clades, lca ):
        tails = []
        for c in clades:
            l = self.get_lineage( c )
            l.reverse( )
            t = []
            for c2 in l:
                if c2 == lca:
                    break
                else:
                    t.append( c2 )
            t.reverse( )
            tails.append( t )
        return tails

    def get_sisters( self, clade ):
        ret = set( )
        parent = self.get_parent( clade )
        for child in self.get_children( parent ):
            if child != clade:
                ret.add( child )
        return ret

    def get_leaf_count( self, clade ):
        ret = None
        if clade in self.known_leaf_count:
            ret = self.known_leaf_count[clade]
        elif clade not in self.children:
            ret = 1
        else:
            ret = 0
            for c in self.children[clade]:
                ret += self.get_leaf_count( c )
        self.known_leaf_count[clade] = ret
        return ret

# ---------------------------------------------------------------
# ---------------------------------------------------------------
# WORKING WITH INTERVALS
# ---------------------------------------------------------------
# ---------------------------------------------------------------

class INode:

    """interval node: represents an interval + some network properties""" 

    def __init__( self, start, stop, strand="+" ):
        self.start, self.stop = sorted( [start, stop] )
        self.strand = strand
        self.neighbors = set()
        self.visited = False

    def __len__( self ):
        return self.stop - self.start + 1

    def attach( self, node ):
        self.neighbors.update( [node] )

    def get_connected_component( self ):
        # modified to use breadth-first search
        cc, front = {self}, {self}
        while any( [not inode.visited for inode in front] ):
            new_front = set()
            for inode in front:
                if not inode.visited:
                    inode.visited = True
                    new_front.update( inode.neighbors )
            cc.update( new_front )
            front = new_front
        return list( cc )

    def to_list( self ):
        return [self.start, self.stop, self.strand] 

def calc_overlap( a1, a2, b1, b2, normalize=True ):
    """ compute overlap between two intervals """
    overlap = None
    a1, a2 = sorted( [a1, a2] )
    b1, b2 = sorted( [b1, b2] )
    if b1 > a2 or a1 > b2:
        overlap = 0
    else:
        outleft, inleft, inright, outright = sorted( [a1, a2, b1, b2] )
        overlap = inright - inleft + 1
        if normalize:
            denom = min( (a2 - a1 + 1), (b2 - b1 + 1) )
            overlap /= float( denom )
    return overlap

# ---------------------------------------------------------------
# ---------------------------------------------------------------
# WORKING WITH SAM OUTPUT
# ---------------------------------------------------------------
# ---------------------------------------------------------------

"""
@HD  VN:1.0                SO:unsorted
@SQ  SN:SRS011061_k119_3   LN:961
@SQ  SN:SRS011061_k119_5   LN:837
@SQ  SN:SRS011061_k119_11  LN:502
...
61NLYAAXX100508:5:100:10002:9010   83   SRS011061_k119_37319  25069  42  101M      =  25052  -118
61NLYAAXX100508:5:100:10002:9010   163  SRS011061_k119_37319  25052  42  46M1I43M  =  25069  118
61NLYAAXX100508:5:100:10002:9075   83   SRS011061_k119_14610  17113  42  101M      =  16942  -272
61NLYAAXX100508:5:100:10002:9075   163  SRS011061_k119_14610  16942  42  46M1I43M  =  17113  272
61NLYAAXX100508:5:100:10003:17250  77   *                     0      0   *         *  0      0
61NLYAAXX100508:5:100:10003:17250  141  *                     0      0   *         *  0      0
61NLYAAXX100508:5:100:10003:3146   99   SRS011061_k119_83764  304    42  101M      =  366    152
61NLYAAXX100508:5:100:10003:3146   147  SRS011061_k119_83764  366    42  90M       =  304    -152
"""

class SAMHit:

    """ Some data about an aligned read in a SAM file """

    def __init__( self, sam_row ):

        self.qseqid  = sam_row[0]
        self.sseqid  = sam_row[2]
        self.sstart  = int( sam_row[3] )
        self.send    = self.sstart + cigar_length( sam_row[5] ) - 1

def cigar_length( cigar ):
    counts = [int( c ) for c in re.split( "[A-Z]+", cigar ) if c != ""]
    sigils = [s        for s in re.split( "[0-9]+", cigar ) if s != ""]
    # ignore read-only bands
    return sum( [c for c, s in zip( counts, sigils ) if s in "DHMNSX="] )

def iter_sam_hits( sam_file ):
    with try_open( sam_file ) as fh:
        for row in csv.reader( fh, dialect="excel-tab" ):
            # header
            if row[0][0] == "@":
                continue
            # mandatory fields for a hit line
            elif len( row ) < 11:
                continue
            # aligned read
            elif row[2] != "*":
                yield SAMHit( row )

# ---------------------------------------------------------------
# ---------------------------------------------------------------
# TESTING
# ---------------------------------------------------------------
# ---------------------------------------------------------------

if __name__ == "__main__":
    pass



import setuptools

with open( "README.md" ) as fh:
    long_description = fh.read( )

setuptools.setup(
    name = "waafle",
    version = "0.1.0dev",
    author = "Eric Franzosa",
    author_email = "franzosa@hsph.harvard.edu",
    license = "MIT",
    description = "WAAFLE: a Workflow to Annotate Assemblies and find LGT Events",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://huttenhower.sph.harvard.edu/waafle",
    packages = setuptools.find_packages( ),
    classifiers = [
        "Programming Language :: Python",
    ],
    entry_points = {
        "console_scripts": [
            "waafle_search = waafle.waafle_search:main",
            "waafle_genecaller = waafle.waafle_genecaller:main",
            "waafle_orgscorer = waafle.waafle_orgscorer:main",
            "waafle_junctions = waafle.waafle_junctions:main",
            "waafle_qc = waafle.waafle_qc:main",
        ],
    },
    install_requires = [
        "numpy >= 1.13.0",
    ],
)



#!/usr/bin/env python

"""
This module is a part of:
WAAFLE, a [W]orkflow to [A]nnotate [A]ssemblies and [F]ind [L]GT [E]vents

Copyright (c) 2019 Harvard T.H. Chan School of Public Health

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function # Python 2.7+ required
import os
import sys
import csv
import argparse
import re
from collections import Counter

from waafle import utils as wu

# ---------------------------------------------------------------
# description
# ---------------------------------------------------------------

description = wu.describe( """
{SCRIPT}: Applies junction results to QC WAAFLE calls.

Filter WAAFLE lgt output to require that junctions
be supported by mate-pair or coverage evidence.
""" )

# ---------------------------------------------------------------
# output formats
# ---------------------------------------------------------------

c_formats = {}
for name, items in c_formats.items( ):
    c_formats[name] = [k for k in items.split( "\n" ) if k != ""]

# ---------------------------------------------------------------
# cli
# ---------------------------------------------------------------

def get_args( ):

    parser = argparse.ArgumentParser(
        description=description, 
        formatter_class=argparse.RawTextHelpFormatter,
        )

    g = parser.add_argument_group( "required inputs" )
    g.add_argument( 
        "contig_profile",
        help="lgt output from waafle_orgscorer (tsv format)",
        )
    g.add_argument( 
        "junctions",
        help="output from waafle_junctions for contigs of interest",
        )

    g = parser.add_argument_group( "filtering parameters" )
    g.add_argument( 
        "--min-junction-hits",
        type=int,
        default=2,
        metavar="<int>",
        help="minimum read-hits to 'ok' a junction\n[default: 2]",
        )
    g.add_argument( 
        "--min-junction-ratio",
        type=int,
        default=0.5,
        metavar="<float>",
        help="minimum coverage (relative to flanking genes) to 'ok' a junction\n[default: 0.5]",
        )
    """
    g.add_argument( 
        "--min-contig-genes",
        type=int,
        default=2,
        metavar="<int>",
        help="minimum gene count for the contig\n[default: 1]",
        )
    g.add_argument( 
        "--min-contig-length",
        type=int,
        default=500,
        metavar="<int>",
        help="minimum length for the contig\n[default: 500]",
        )
    """

    g = parser.add_argument_group( "misc options" )
    """
    g.add_argument( 
        "--ab-only",
        action="store_true",
        help="Only apply filtering to LGT junctions (AB/BA)",
        )
    """
    g.add_argument( 
        "--outfile",
        type=str,
        default=None,
        metavar="<path>",
        help="Path for filtered outputs\n[default: derive from input]",
        )

    args = parser.parse_args( )
    return args

# ---------------------------------------------------------------
# main
# ---------------------------------------------------------------

def main( ):

    args = get_args( )

    # load junctions data
    hits = {}
    covs = {}
    wu.say( "Loading junctions report." )
    F = wu.Frame( args.junctions )
    # loop over junctions
    for R in F.iter_rowdicts( ):
        contig = R["CONTIG"]
        gene1 = R["GENE1"]
        gene2 = R["GENE2"]
        hits.setdefault( contig, {} )[(gene1, gene2)] = int( R["JUNCTION_HITS"] )
        covs.setdefault( contig, {} )[(gene1, gene2)] = float( R["RATIO"] )

    # filter contigs
    total = 0
    failed = 0
    outfile = args.outfile
    if outfile is None:
        outfile = args.contig_profile + ".qc_pass"
    # load results, open new file, write headers
    F = wu.Frame( args.contig_profile )
    fh = wu.try_open( outfile, "w" )
    wu.write_rowdict( None, F.headers, file=fh )
    # loop over contigs
    for R in F.iter_rowdicts( ):
        total += 1
        contig = R["CONTIG_NAME"]
        # contig-level filters
        if contig not in hits or contig not in covs:
            failed += 1
            wu.say( "Missing junction data for contig:", contig )
            continue        
        loci = R["LOCI"].split( "|" )
        synteny = R["SYNTENY"]
        qc_pass = True
        for i in range( len( loci ) - 1 ):
            my_test = True
            spair = synteny[i] + synteny[i+1]
            if spair not in ["AB", "BA"]:
                continue
            gpair = (loci[i], loci[i+1])
            my_hits = hits[contig].get( gpair, -1 )
            my_hits = my_hits >= args.min_junction_hits
            my_covs = covs[contig].get( gpair, -1 )
            my_covs = my_covs >= args.min_junction_ratio
            my_test = my_hits or my_covs
            qc_pass = qc_pass and my_test
        if not qc_pass:
            failed += 1
            wu.say( "Failed QC:", contig )
        else:
            wu.write_rowdict( R, F.headers, file=fh )

    # wrap-up
    wu.say( "Failure rate: {} of {} ({:.1f}%)".format( failed, total, 100 * failed / float( total ) ) )
    wu.say( "Finished successfully." )           
                
if __name__ == "__main__":
    main( )
